library(Biostrings)
library(parallel)

parse_data <- function(file) {
  reversed_sequences <- Biostrings::readQualityScaledDNAStringSet(file)
  sequences <- Biostrings::reverseComplement(reversed_sequences)
  vj_segments <- union(
    readRDS("data/v_segments.rds"),
    readRDS("data/j_segments_phe.rds")
  )
  return(list(sequences, vj_segments))
}

parse_metadata <- function(metadata) {
  id_elements <- unlist(strsplit(metadata, split = " "))
  v_identifier <- id_elements[2]
  j_identifier <- id_elements[3]
  return(list(v_id = v_identifier, j_id = j_identifier))
}

match_id_sequence <- function(names, vdj_segments, id) {
  matches <- grep(names, pattern = id)
  row <- matches[1]
  return(as.character(vdj_segments[row]))
}

get_vj_sequence <- function(metadata, names, vdj_segments) {
  identifiers <- parse_metadata(metadata)
  v_sequence <- match_id_sequence(names, vdj_segments, id = identifiers["v_id"])
  j_sequence <- match_id_sequence(names, vdj_segments, id = identifiers["j_id"])
  return(list(v_seq = v_sequence, j_seq = j_sequence))
}

fetch_vj_sequences <- function(sequences, vdj_segments) {
  vj_sequences <- sapply(names(sequences),
    names(vdj_segments),
    vdj_segments,
    FUN = get_vj_sequence
  )
  results <- data.frame(t(vj_sequences))
  return(results)
}

align_sequence <- function(sequence, vdj_segment) {
  return(Biostrings::pairwiseAlignment(
    subject = sequence,
    pattern = vdj_segment,
    type = "global-local",
    gapOpening = 1
  ))
}

locate_cysteine <- function(sequence, v_segment) {
  codons <- Biostrings::DNAStringSet(c("TGT", "TGC"))
  matches <- Biostrings::matchPDict(
    pdict = codons,
    subject = DNAString(toString(v_segment))
  )
  position <- as.data.frame(intersect(matches[[1]], matches[[2]]))
  return(position)
}

# TODO Extract CDR3
get_hvr_sequences <- function(sequences, vdj_segments) {
  df <- fetch_vj_sequences(sequences, vdj_segments)
  v_alignment <- parallel::mcmapply(sequences, df$v_seq, FUN = align_sequence)
  hvr_start <- parallel::mcmapply(sequences, v_alignment, FUN = locate_cysteine)
  hvr_start_df <- as.data.frame(t(hvr_start))
}

data <- parse_data(file = "data/curesim_sequence.fastq")
hvr_sequences <- get_hvr_sequences(
  sequences = data[[1]],
  vdj_segments = data[[2]]
)
